@@ -10,21 +10,21 @@ class TestWEBrickHTTPRequest < Test::Unit::TestCase
   end
 
   def test_simple_request
-    msg = <<-_end_of_message_
-GET /
-    _end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET /
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     req.parse(StringIO.new(msg))
     assert(req.meta_vars) # fails if @header was not initialized and iteration is attempted on the nil reference
   end
 
   def test_parse_09
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /
       foobar    # HTTP/0.9 request don't have header nor entity body.
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal("GET", req.request_method)
     assert_equal("/", req.unparsed_uri)
     assert_equal(WEBrick::HTTPVersion.new("0.9"), req.http_version)
@@ -36,12 +36,12 @@ GET /
   end
 
   def test_parse_10
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET / HTTP/1.0
 
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal("GET", req.request_method)
     assert_equal("/", req.unparsed_uri)
     assert_equal(WEBrick::HTTPVersion.new("1.0"), req.http_version)
@@ -53,12 +53,12 @@ GET /
   end
 
   def test_parse_11
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /path HTTP/1.1
 
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal("GET", req.request_method)
     assert_equal("/path", req.unparsed_uri)
     assert_equal("", req.script_name)
@@ -72,17 +72,173 @@ GET /
   end
 
   def test_request_uri_too_large
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /#{"a"*2084} HTTP/1.1
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     assert_raise(WEBrick::HTTPStatus::RequestURITooLarge){
-      req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+      req.parse(StringIO.new(msg))
+    }
+  end
+
+  def test_invalid_content_length_header
+    ['', ' ', ' +1', ' -1', ' a'].each do |cl|
+      msg = <<~HTTP.gsub("\n", "\r\n")
+        GET / HTTP/1.1
+        Content-Length:#{cl}
+
+      HTTP
+      req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+      assert_raise(WEBrick::HTTPStatus::BadRequest){
+        req.parse(StringIO.new(msg))
+      }
+    end
+  end
+
+  def test_bare_lf_request_line
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET / HTTP/1.1
+      Content-Length: 0\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::EOFError){
+      req.parse(StringIO.new(msg))
+    }
+  end
+
+  def test_bare_lf_header
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET / HTTP/1.1\r
+      Content-Length: 0
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::BadRequest){
+      req.parse(StringIO.new(msg))
+    }
+  end
+
+  def test_header_vt_ff_whitespace
+    msg = <<~HTTP
+      GET / HTTP/1.1\r
+      Foo: \x0b1\x0c\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    req.parse(StringIO.new(msg))
+    assert_equal("\x0b1\x0c", req["Foo"])
+
+    msg = <<~HTTP
+      GET / HTTP/1.1\r
+      Foo: \x0b1\x0c\r
+       \x0b2\x0c\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    req.parse(StringIO.new(msg))
+    assert_equal("\x0b1\x0c \x0b2\x0c", req["Foo"])
+  end
+
+  def test_bare_cr_request_line
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET / HTTP/1.1\r\r
+      Content-Length: 0\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::BadRequest){
+      req.parse(StringIO.new(msg))
+    }
+  end
+
+  def test_bare_cr_header
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET / HTTP/1.1\r
+      Content-Type: foo\rbar\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::BadRequest){
+      req.parse(StringIO.new(msg))
+    }
+  end
+
+  def test_invalid_request_lines
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET  / HTTP/1.1\r
+      Content-Length: 0\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::BadRequest){
+      req.parse(StringIO.new(msg))
+    }
+
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET /  HTTP/1.1\r
+      Content-Length: 0\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::BadRequest){
+      req.parse(StringIO.new(msg))
+    }
+
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET /\r HTTP/1.1\r
+      Content-Length: 0\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::BadRequest){
+      req.parse(StringIO.new(msg))
+    }
+
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET / HTTP/1.1 \r
+      Content-Length: 0\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::BadRequest){
+      req.parse(StringIO.new(msg))
+    }
+  end
+
+  def test_duplicate_content_length_header
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET / HTTP/1.1
+      Content-Length: 1
+      Content-Length: 2
+
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::BadRequest){
+      req.parse(StringIO.new(msg))
+    }
+  end
+
+  def test_content_length_and_transfer_encoding_headers_smuggling
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      POST /user HTTP/1.1
+      Content-Length: 28
+      Transfer-Encoding: chunked
+
+      0
+
+      GET /admin HTTP/1.1
+
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    req.parse(StringIO.new(msg))
+    assert_raise(WEBrick::HTTPStatus::BadRequest){
+      req.body
     }
   end
 
   def test_parse_headers
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /path HTTP/1.1
       Host: test.ruby-lang.org:8080
       Connection: close
@@ -93,13 +249,13 @@ GET /
       Accept-Language: en;q=0.5, *; q=0
       Accept-Language: ja
       Content-Type: text/plain
-      Content-Length: 7
+      Content-Length: 8
       X-Empty-Header:
 
       foobar
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal(
       URI.parse("http://test.ruby-lang.org:8080/path"), req.request_uri)
     assert_equal("test.ruby-lang.org", req.host)
@@ -110,88 +266,88 @@ GET /
       req.accept)
     assert_equal(%w(gzip compress identity *), req.accept_encoding)
     assert_equal(%w(ja en *), req.accept_language)
-    assert_equal(7, req.content_length)
+    assert_equal(8, req.content_length)
     assert_equal("text/plain", req.content_type)
-    assert_equal("foobar\n", req.body)
+    assert_equal("foobar\r\n", req.body)
     assert_equal("", req["x-empty-header"])
     assert_equal(nil, req["x-no-header"])
     assert(req.query.empty?)
   end
 
   def test_parse_header2()
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       POST /foo/bar/../baz?q=a HTTP/1.0
-      Content-Length: 9
+      Content-Length: 10
       User-Agent:
         FOO   BAR
         BAZ
 
       hogehoge
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal("POST", req.request_method)
     assert_equal("/foo/baz", req.path)
     assert_equal("", req.script_name)
     assert_equal("/foo/baz", req.path_info)
-    assert_equal("9", req['content-length'])
+    assert_equal("10", req['content-length'])
     assert_equal("FOO   BAR BAZ", req['user-agent'])
-    assert_equal("hogehoge\n", req.body)
+    assert_equal("hogehoge\r\n", req.body)
   end
 
   def test_parse_headers3
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /path HTTP/1.1
       Host: test.ruby-lang.org
 
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal(URI.parse("http://test.ruby-lang.org/path"), req.request_uri)
     assert_equal("test.ruby-lang.org", req.host)
     assert_equal(80, req.port)
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /path HTTP/1.1
       Host: 192.168.1.1
 
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal(URI.parse("http://192.168.1.1/path"), req.request_uri)
     assert_equal("192.168.1.1", req.host)
     assert_equal(80, req.port)
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /path HTTP/1.1
       Host: [fe80::208:dff:feef:98c7]
 
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal(URI.parse("http://[fe80::208:dff:feef:98c7]/path"),
                  req.request_uri)
     assert_equal("[fe80::208:dff:feef:98c7]", req.host)
     assert_equal(80, req.port)
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /path HTTP/1.1
       Host: 192.168.1.1:8080
 
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal(URI.parse("http://192.168.1.1:8080/path"), req.request_uri)
     assert_equal("192.168.1.1", req.host)
     assert_equal(8080, req.port)
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /path HTTP/1.1
       Host: [fe80::208:dff:feef:98c7]:8080
 
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     assert_equal(URI.parse("http://[fe80::208:dff:feef:98c7]:8080/path"),
                  req.request_uri)
     assert_equal("[fe80::208:dff:feef:98c7]", req.host)
@@ -200,13 +356,13 @@ GET /
 
   def test_parse_get_params
     param = "foo=1;foo=2;foo=3;bar=x"
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /path?#{param} HTTP/1.1
       Host: test.ruby-lang.org:8080
 
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     query = req.query
     assert_equal("1", query["foo"])
     assert_equal(["1", "2", "3"], query["foo"].to_ary)
@@ -217,16 +373,16 @@ GET /
 
   def test_parse_post_params
     param = "foo=1;foo=2;foo=3;bar=x"
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       POST /path?foo=x;foo=y;foo=z;bar=1 HTTP/1.1
       Host: test.ruby-lang.org:8080
       Content-Length: #{param.size}
       Content-Type: application/x-www-form-urlencoded
 
       #{param}
-    _end_of_message_
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-    req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+    req.parse(StringIO.new(msg))
     query = req.query
     assert_equal("1", query["foo"])
     assert_equal(["1", "2", "3"], query["foo"].to_ary)
@@ -238,13 +394,12 @@ GET /
   def test_chunked
     crlf = "\x0d\x0a"
     expect = File.binread(__FILE__).freeze
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       POST /path HTTP/1.1
       Host: test.ruby-lang.org:8080
       Transfer-Encoding: chunked
 
-    _end_of_message_
-    msg.gsub!(/^ {6}/, "")
+    HTTP
     File.open(__FILE__){|io|
       while chunk = io.read(100)
         msg << chunk.size.to_s(16) << crlf
@@ -264,8 +419,63 @@ GET /
     assert_equal(expect, dst.string)
   end
 
+  def test_bad_chunked
+    msg = <<~HTTP
+      POST /path HTTP/1.1\r
+      Transfer-Encoding: chunked\r
+      \r
+      01x1\r
+      \r
+      1
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    req.parse(StringIO.new(msg))
+    assert_raise(WEBrick::HTTPStatus::BadRequest){ req.body }
+
+    # chunked req.body_reader
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    req.parse(StringIO.new(msg))
+    dst = StringIO.new
+    assert_raise(WEBrick::HTTPStatus::BadRequest) do
+      IO.copy_stream(req.body_reader, dst)
+    end
+  end
+
+  def test_bad_chunked_extra_data
+    msg = <<~HTTP
+      POST /path HTTP/1.1\r
+      Transfer-Encoding: chunked\r
+      \r
+      3\r
+      ABCthis-all-gets-ignored\r
+      0\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    req.parse(StringIO.new(msg))
+    assert_raise(WEBrick::HTTPStatus::BadRequest){ req.body }
+
+    # chunked req.body_reader
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    req.parse(StringIO.new(msg))
+    dst = StringIO.new
+    assert_raise(WEBrick::HTTPStatus::BadRequest) do
+      IO.copy_stream(req.body_reader, dst)
+    end
+  end
+
+  def test_null_byte_in_header
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      POST /path HTTP/1.1\r
+      Evil: evil\x00\r
+      \r
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::BadRequest){ req.parse(StringIO.new(msg)) }
+  end
+
   def test_forwarded
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /foo HTTP/1.1
       Host: localhost:10080
       User-Agent: w3m/0.5.2
@@ -274,8 +484,7 @@ GET /
       X-Forwarded-Server: server.example.com
       Connection: Keep-Alive
 
-    _end_of_message_
-    msg.gsub!(/^ {6}/, "")
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     req.parse(StringIO.new(msg))
     assert_equal("server.example.com", req.server_name)
@@ -285,7 +494,7 @@ GET /
     assert_equal("123.123.123.123", req.remote_ip)
     assert(!req.ssl?)
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /foo HTTP/1.1
       Host: localhost:10080
       User-Agent: w3m/0.5.2
@@ -294,8 +503,7 @@ GET /
       X-Forwarded-Server: server.example.com
       Connection: Keep-Alive
 
-    _end_of_message_
-    msg.gsub!(/^ {6}/, "")
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     req.parse(StringIO.new(msg))
     assert_equal("server.example.com", req.server_name)
@@ -305,7 +513,7 @@ GET /
     assert_equal("123.123.123.123", req.remote_ip)
     assert(!req.ssl?)
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /foo HTTP/1.1
       Host: localhost:10080
       Client-IP: 234.234.234.234
@@ -316,8 +524,7 @@ GET /
       X-Requested-With: XMLHttpRequest
       Connection: Keep-Alive
 
-    _end_of_message_
-    msg.gsub!(/^ {6}/, "")
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     req.parse(StringIO.new(msg))
     assert_equal("server.example.com", req.server_name)
@@ -327,7 +534,7 @@ GET /
     assert_equal("234.234.234.234", req.remote_ip)
     assert(req.ssl?)
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /foo HTTP/1.1
       Host: localhost:10080
       Client-IP: 234.234.234.234
@@ -338,8 +545,7 @@ GET /
       X-Requested-With: XMLHttpRequest
       Connection: Keep-Alive
 
-    _end_of_message_
-    msg.gsub!(/^ {6}/, "")
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     req.parse(StringIO.new(msg))
     assert_equal("server1.example.com", req.server_name)
@@ -349,7 +555,7 @@ GET /
     assert_equal("234.234.234.234", req.remote_ip)
     assert(req.ssl?)
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /foo HTTP/1.1
       Host: localhost:10080
       Client-IP: 234.234.234.234
@@ -360,8 +566,7 @@ GET /
       X-Requested-With: XMLHttpRequest
       Connection: Keep-Alive
 
-    _end_of_message_
-    msg.gsub!(/^ {6}/, "")
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     req.parse(StringIO.new(msg))
     assert_equal("server1.example.com", req.server_name)
@@ -371,7 +576,7 @@ GET /
     assert_equal("234.234.234.234", req.remote_ip)
     assert(req.ssl?)
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       GET /foo HTTP/1.1
       Host: localhost:10080
       Client-IP: 234.234.234.234
@@ -382,8 +587,7 @@ GET /
       X-Requested-With: XMLHttpRequest
       Connection: Keep-Alive
 
-    _end_of_message_
-    msg.gsub!(/^ {6}/, "")
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     req.parse(StringIO.new(msg))
     assert_equal("server1.example.com", req.server_name)
@@ -395,12 +599,11 @@ GET /
   end
 
   def test_continue_sent
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       POST /path HTTP/1.1
       Expect: 100-continue
 
-    _end_of_message_
-    msg.gsub!(/^ {6}/, "")
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     req.parse(StringIO.new(msg))
     assert req['expect']
@@ -412,11 +615,10 @@ GET /
   end
 
   def test_continue_not_sent
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       POST /path HTTP/1.1
 
-    _end_of_message_
-    msg.gsub!(/^ {6}/, "")
+    HTTP
     req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
     req.parse(StringIO.new(msg))
     assert !req['expect']
@@ -427,42 +629,42 @@ GET /
 
   def test_bad_messages
     param = "foo=1;foo=2;foo=3;bar=x"
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       POST /path?foo=x;foo=y;foo=z;bar=1 HTTP/1.1
       Host: test.ruby-lang.org:8080
       Content-Type: application/x-www-form-urlencoded
 
       #{param}
-    _end_of_message_
+    HTTP
     assert_raise(WEBrick::HTTPStatus::LengthRequired){
       req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-      req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+      req.parse(StringIO.new(msg))
       req.body
     }
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       POST /path?foo=x;foo=y;foo=z;bar=1 HTTP/1.1
       Host: test.ruby-lang.org:8080
       Content-Length: 100000
 
       body is too short.
-    _end_of_message_
+    HTTP
     assert_raise(WEBrick::HTTPStatus::BadRequest){
       req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-      req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+      req.parse(StringIO.new(msg))
       req.body
     }
 
-    msg = <<-_end_of_message_
+    msg = <<~HTTP.gsub("\n", "\r\n")
       POST /path?foo=x;foo=y;foo=z;bar=1 HTTP/1.1
       Host: test.ruby-lang.org:8080
       Transfer-Encoding: foobar
 
       body is too short.
-    _end_of_message_
+    HTTP
     assert_raise(WEBrick::HTTPStatus::NotImplemented){
       req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
-      req.parse(StringIO.new(msg.gsub(/^ {6}/, "")))
+      req.parse(StringIO.new(msg))
       req.body
     }
   end
@@ -473,4 +675,30 @@ GET /
       req.parse(StringIO.new(""))
     }
   end
+
+  def test_eof_raised_with_missing_line_between_headers_and_body
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET / HTTP/1.0
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::EOFError) {
+      req.parse(StringIO.new(msg))
+    }
+
+    msg = <<~HTTP.gsub("\n", "\r\n")
+      GET / HTTP/1.0
+      Foo: 1
+    HTTP
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    assert_raise(WEBrick::HTTPStatus::EOFError) {
+      req.parse(StringIO.new(msg))
+    }
+  end
+
+  def test_cookie_join
+    req = WEBrick::HTTPRequest.new(WEBrick::Config::HTTP)
+    req.parse(StringIO.new("GET / HTTP/1.1\r\ncookie: a=1\r\ncookie: b=2\r\n\r\n"))
+    assert_equal 2, req.cookies.length
+    assert_equal 'a=1; b=2', req['cookie']
+  end
 end
